/*-
 * Copyright (C) 2012 Christian Hesse
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 */

#include <config.h>

#include <gio/gio.h>
#include <polkit/polkit.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <libxfsm/xfsm-util.h>
#include <xfce4-session/xfsm-systemd.h>



#define SYSTEMD_DBUS_NAME               "org.freedesktop.login1"
#define SYSTEMD_DBUS_PATH               "/org/freedesktop/login1"
#define SYSTEMD_DBUS_INTERFACE          "org.freedesktop.login1.Manager"
#define SYSTEMD_REBOOT_ACTION           "Reboot"
#define SYSTEMD_POWEROFF_ACTION         "PowerOff"
#define SYSTEMD_SUSPEND_ACTION          "Suspend"
#define SYSTEMD_HIBERNATE_ACTION        "Hibernate"
#define SYSTEMD_AUTH_REBOOT             "org.freedesktop.login1.reboot"
#define SYSTEMD_AUTH_POWEROFF           "org.freedesktop.login1.power-off"
#define SYSTEMD_AUTH_SUSPEND            "org.freedesktop.login1.suspend"
#define SYSTEMD_AUTH_HIBERNATE          "org.freedesktop.login1.hibernate"
#define SYSTEMD_CAN_REBOOT              "CanReboot"
#define SYSTEMD_CAN_POWEROFF            "CanPowerOff"
#define SYSTEMD_CAN_SUSPEND             "CanSuspend"
#define SYSTEMD_CAN_HIBERNATE           "CanHibernate"



static void     xfsm_systemd_finalize     (GObject         *object);
static gboolean xfsm_systemd_proxy_ensure (XfsmSystemd     *systemd,
                                           GError         **error);
static void     xfsm_systemd_proxy_free   (XfsmSystemd     *systemd);


struct _XfsmSystemdClass
{
  GObjectClass __parent__;
};

struct _XfsmSystemd
{
  GObject __parent__;

  DBusGConnection *dbus_conn;
  DBusGProxy      *systemd_proxy;
  PolkitAuthority *authority;
  PolkitSubject   *subject;
};



G_DEFINE_TYPE (XfsmSystemd, xfsm_systemd, G_TYPE_OBJECT)



static void
xfsm_systemd_class_init (XfsmSystemdClass *klass)
{
  GObjectClass *gobject_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = xfsm_systemd_finalize;
}



static void
xfsm_systemd_init (XfsmSystemd *systemd)
{
  systemd->authority = polkit_authority_get_sync (NULL, NULL);
  systemd->subject = polkit_unix_process_new (getpid());
}



static void
xfsm_systemd_finalize (GObject *object)
{
  XfsmSystemd *systemd = XFSM_SYSTEMD (object);

  g_object_unref (G_OBJECT (systemd->authority));
  g_object_unref (G_OBJECT (systemd->subject));
  xfsm_systemd_proxy_free (systemd);

  (*G_OBJECT_CLASS (xfsm_systemd_parent_class)->finalize) (object);
}



static gboolean
xfsm_systemd_lock_screen (GError **error)
{
  XfconfChannel *channel;
  gboolean       ret = TRUE;

  channel = xfsm_open_config ();
  if (xfconf_channel_get_bool (channel, "/shutdown/LockScreen", FALSE))
      ret = g_spawn_command_line_async ("xflock4", error);

  return ret;
}



static gboolean
xfsm_systemd_proxy_ensure (XfsmSystemd *systemd,
                          GError      **error)
{
  GError         *local_error = NULL;
  DBusConnection *connection;

  if (systemd->dbus_conn == NULL)
    {
      systemd->dbus_conn = dbus_g_bus_get (DBUS_BUS_SYSTEM, &local_error);
      if (systemd->dbus_conn == NULL)
        goto error1;

      connection = dbus_g_connection_get_connection (systemd->dbus_conn);
      dbus_connection_set_exit_on_disconnect (connection, FALSE);
    }

  if (systemd->systemd_proxy == NULL)
    {
      systemd->systemd_proxy = dbus_g_proxy_new_for_name (systemd->dbus_conn,
                                                          SYSTEMD_DBUS_NAME,
                                                          SYSTEMD_DBUS_PATH,
                                                          SYSTEMD_DBUS_INTERFACE);
      if (systemd->systemd_proxy == NULL)
        {
          g_set_error (&local_error, DBUS_GERROR, DBUS_GERROR_FAILED,
                       "Failed to get proxy for %s",
                       SYSTEMD_DBUS_NAME);
          goto error1;
        }
    }

  return TRUE;

  error1:

  g_propagate_error (error, local_error);
  xfsm_systemd_proxy_free (systemd);

  return FALSE;
}



static void
xfsm_systemd_proxy_free (XfsmSystemd *systemd)
{
  if (systemd->systemd_proxy != NULL)
    {
      g_object_unref (G_OBJECT (systemd->systemd_proxy));
      systemd->systemd_proxy = NULL;
    }

  if (systemd->dbus_conn != NULL)
    {
      dbus_g_connection_unref (systemd->dbus_conn);
      systemd->dbus_conn = NULL;
    }
}



static gboolean
xfsm_systemd_can_method (XfsmSystemd  *systemd,
                         const gchar  *can_property, /* if the system is capable */
                         gboolean     *can_return,
                         const gchar  *auth_method, /* if the user is allowed */
                         gboolean     *auth_return,
                         GError      **error)
{
  PolkitAuthorizationResult *res;
  char                      *retstr = NULL;

  *can_return = FALSE;
  *auth_return = FALSE;

  if (G_UNLIKELY (!xfsm_systemd_proxy_ensure (systemd, error)))
    return FALSE;

  if (!dbus_g_proxy_call (systemd->systemd_proxy, can_property, error,
                          G_TYPE_INVALID,
                          G_TYPE_STRING, &retstr,
                          G_TYPE_INVALID))
    return FALSE;

  if (retstr && !strcmp (retstr, "yes"))
    *can_return = TRUE;
  g_free (retstr);

  if (*can_return)
    {
      res = polkit_authority_check_authorization_sync (systemd->authority,
                                                       systemd->subject,
                                                       auth_method,
                                                       NULL,
                                                       POLKIT_CHECK_AUTHORIZATION_FLAGS_NONE,
                                                       NULL,
                                                       error);

      if (res == NULL)
        return FALSE;

      *auth_return = polkit_authorization_result_get_is_authorized (res)
                    || polkit_authorization_result_get_is_challenge (res);

      g_object_unref (G_OBJECT (res));
    }

  return TRUE;
}



static gboolean
xfsm_systemd_try_method (XfsmSystemd  *systemd,
                         const gchar  *method,
                         GError      **error)
{
  return dbus_g_proxy_call (systemd->systemd_proxy, method, error,
                            G_TYPE_BOOLEAN, TRUE,
                            G_TYPE_INVALID,
                            G_TYPE_INVALID);
}



XfsmSystemd *
xfsm_systemd_get (void)
{
  static XfsmSystemd *object = NULL;

  if (G_LIKELY (object != NULL))
    {
      g_object_ref (G_OBJECT (object));
    }
  else
    {
      object = g_object_new (XFSM_TYPE_SYSTEMD, NULL);
      g_object_add_weak_pointer (G_OBJECT (object), (gpointer) &object);
    }

  return object;
}



gboolean
xfsm_systemd_try_restart (XfsmSystemd  *systemd,
                          GError      **error)
{
  return xfsm_systemd_try_method (systemd,
                                  SYSTEMD_REBOOT_ACTION,
                                  error);
}



gboolean
xfsm_systemd_try_shutdown (XfsmSystemd  *systemd,
                           GError      **error)
{
  return xfsm_systemd_try_method (systemd,
                                  SYSTEMD_POWEROFF_ACTION,
                                  error);
}



gboolean
xfsm_systemd_try_suspend (XfsmSystemd  *systemd,
                          GError      **error)
{
  if (!xfsm_systemd_lock_screen (error))
    return FALSE;

  return xfsm_systemd_try_method (systemd,
                                  SYSTEMD_SUSPEND_ACTION,
                                  error);
}



gboolean
xfsm_systemd_try_hibernate (XfsmSystemd  *systemd,
                            GError      **error)
{
  if (!xfsm_systemd_lock_screen (error))
    return FALSE;

  return xfsm_systemd_try_method (systemd,
                                  SYSTEMD_HIBERNATE_ACTION,
                                  error);
}



gboolean
xfsm_systemd_can_restart (XfsmSystemd  *systemd,
                          gboolean     *can_restart,
                          GError      **error)
{
  gboolean auth_restart = FALSE;
  gboolean ret;

  ret = xfsm_systemd_can_method (systemd,
                                 SYSTEMD_CAN_REBOOT,
                                 can_restart,
                                 SYSTEMD_AUTH_REBOOT,
                                 &auth_restart,
                                 error);

  if (auth_restart == FALSE)
    *can_restart = FALSE;

  return ret;
}



gboolean
xfsm_systemd_can_shutdown (XfsmSystemd  *systemd,
                           gboolean     *can_shutdown,
                           GError      **error)
{
  gboolean auth_shutdown = FALSE;
  gboolean ret;

  ret = xfsm_systemd_can_method (systemd,
                                 SYSTEMD_CAN_POWEROFF,
                                 can_shutdown,
                                 SYSTEMD_AUTH_POWEROFF,
                                 &auth_shutdown,
                                 error);

  if (auth_shutdown == FALSE)
    *can_shutdown = FALSE;

  return ret;
}



gboolean
xfsm_systemd_can_suspend (XfsmSystemd  *systemd,
                          gboolean     *can_suspend,
                          gboolean     *auth_suspend,
                          GError      **error)
{
  return xfsm_systemd_can_method (systemd,
                                  SYSTEMD_CAN_SUSPEND,
                                  can_suspend,
                                  SYSTEMD_AUTH_SUSPEND,
                                  auth_suspend,
                                  error);
}



gboolean
xfsm_systemd_can_hibernate (XfsmSystemd  *systemd,
                            gboolean     *can_hibernate,
                            gboolean     *auth_hibernate,
                            GError      **error)
{
  return xfsm_systemd_can_method (systemd,
                                  SYSTEMD_CAN_HIBERNATE,
                                  can_hibernate,
                                  SYSTEMD_AUTH_HIBERNATE,
                                  auth_hibernate,
                                  error);
}
